classdef cantilever2DMinComplianceSA
    properties(GetAccess = 'public', SetAccess = 'private')
        myInitialShapeParams; % used for scaling
        myUB; % upper bound on the shape parameters
        myLB;% lower bound on the shape parameters
        myInitialVolume; % used for scaling volume
        myInitialCompliance; % used for scaling compliance
        myInitialStress;
        myNumElems;
        myFiniteDifferenceStepSize;
        myTerminationTolerance;
        myInitialFem; % Initial Fem structure
    end
    methods
        function obj = cantilever2DMinComplianceSA(params0,nElements)
            obj.myNumElems = nElements;
            obj.myInitialShapeParams = params0;
            fem = obj.solveCantileverProblem(params0);
            obj.myInitialFem = fem;
            [obj.myInitialCompliance] = fem.computeCompliance();           
            obj.myInitialVolume = fem.computeArea();
            obj.myInitialStress = fem.myMaxStress;
            obj.myFiniteDifferenceStepSize = 1e-5;
            obj.myTerminationTolerance = 0.01;
        end
        function fem = createCantileverProblem(obj,params)
            cantileverBeam = obj.createGeom(params);
            fem = TriElasticity(cantileverBeam,obj.myNumElems);
            fem = fem.setYoungsModulus(3.05e7);
            fem = fem.setPoissonsRatio(0.28);
            fem = fem.fixEdge([2,10]);
            fem = fem.applyYForceOnEdge(6,-1000);
        end
        function fem = solveCantileverProblem(obj,params)
            fem = obj.createCantileverProblem(params);
            fem = fem.assemble();
            fem = fem.solve();
        end
        function [geom] = createGeom(~,params)
            L = 2.0;%length
            H = 1;%height
            h = 0.2; % length of force edge
            a = params(1); % corner cutouts
            b = params(2); % left edge cutout
            geom.vertices = [b 0;0 -b; 0 -H/2;L-a -H/2;L -H/2+a; L -h/2; L h/2; L H/2-a; L-a H/2; 0 H/2; 0 b ]';
            geom.edges = [1 2; 2 3; 3 4; 4 5; 5 6; 6 7; 7 8; 8 9; 9 10; 10 11; 11 1];
        end
        function obj = setParamsUpperBound(obj,paramsUpper)
            obj.myUB = paramsUpper;
        end
        function obj = setParamsLowerBound(obj,paramsLower)
            obj.myLB = paramsLower;
        end
        function obj = setTerminationTolerance(obj,tol)
            obj.myTerminationTolerance = tol;
        end
        function obj = setFiniteDifferenceStepSize(obj,step)
            obj.myFiniteDifferenceStepSize = step;
        end
        function [f,grad_f] = complianceObjective(obj,x)
            params = x.*obj.myInitialShapeParams;
            fem = obj.solveCantileverProblem(params);
            compliance = fem.computeCompliance();
            f = compliance/obj.myInitialCompliance;
            if (nargout == 2)
                grad_f = complianceSensitivity(obj,fem,x);
            end
        end
        function [cineq,ceq] = volumeConstraint(obj,x)
            params = x.*obj.myInitialShapeParams;
            cantileverBeam = obj.createGeom(params);
            fem = TriElasticity(cantileverBeam,obj.myNumElems);
            currentVolume = fem.computeArea();
            ceq = [];
            cineq = currentVolume/(obj.myInitialVolume) - 1.0;
        end
        function gradJ = complianceSensitivity(obj,fem0,x)
            % The idea is to create a pertrubed geometry, compute its
            % stiffness matrix. Then, take its difference with respect to
            % original stiffness matrix, etc.
            h = obj.myFiniteDifferenceStepSize;
            xh = x + [h 0]; % perturbed the first parameter
            params = xh.*obj.myInitialShapeParams;
            fem = obj.createCantileverProblem(params);
            % project old mesh onto new geometry
            fem = obj.projectMeshOntoBrep(fem0.myMesh,fem);
            fem = fem.assembleK(); % assemble a new stiffness
            KPrime= (fem.myK - fem0.myK)/h; % sensitivity of K
            d = fem0.mySol;
            dJ_da = -d'*KPrime*d;
            dJ_dx1 =  dJ_da*obj.myInitialShapeParams(1);
            
            xh = x +[0 h]; % perturbed the second parameter
            params = xh.*obj.myInitialShapeParams;
            fem = obj.createCantileverProblem(params);
            
            % project old mesh onto new geometry
            fem = obj.projectMeshOntoBrep(fem0.myMesh,fem);
            fem = fem.assembleK(); % assemble        
            KPrime= (fem.myK - fem0.myK)/h; % sensitivity of K
            d = fem0.mySol;
            dJ_db = -d'*KPrime*d;
            dJ_dx2 =  dJ_db*obj.myInitialShapeParams(2);
            gradJ = [dJ_dx1; dJ_dx2];
            gradJ = gradJ/obj.myInitialCompliance;
        end
        function [fem] = projectMeshOntoBrep(~,mesh,fem)
            % project the boundary nodes of a triangle mesh onto a perturbed brep
            newMesh = mesh;% copy over the original mesh
            % find the mesh points that lie on the boundary
            e = mesh.e;
            if (size(mesh.t,1) == 4) % Linear
                bndryNodes = e(1:2,:);
                bndryNodes = unique(bndryNodes(:));
            else % Quadratic
                bndryNodes = e(1:3,:);
                bndryNodes = unique(bndryNodes(:));
            end
            bndryPts = mesh.p(:,bndryNodes);
            [~,closestPts] = fem.distOfPointsToBrep(bndryPts);
            newMesh.p(:,bndryNodes) = closestPts; 
            % overwrite bunch of variables
            fem.myMesh = newMesh;
            fem.myNumNodes = size(newMesh.p,2);
            fem.myNumElems = size(newMesh.t,2);
            fem.myPseudoDensity = ones(fem.myNumElems,1);
        end
        function [results] = optimize(obj)
            LB = obj.myLB./obj.myInitialShapeParams;
            UB = obj.myUB./obj.myInitialShapeParams;
            x0 = ones(1,2);
            opt = optimoptions('fmincon',...
            'TolX',obj.myTerminationTolerance,...
            'TolFun',obj.myTerminationTolerance,...
            'ConstraintTolerance',obj.myTerminationTolerance,...
            'SpecifyObjectiveGradient',true);   
            [xMin,~,flag,output] = fmincon(@obj.complianceObjective,x0,[],[],[],[],LB,UB,@obj.volumeConstraint,opt);
            optimalParams = xMin.*obj.myInitialShapeParams;
            fem = obj.solveCantileverProblem(optimalParams); 
            results.initialParams =  obj.myInitialShapeParams;
            results.initialVolume =  obj.myInitialVolume;
            results.initialCompliance =  obj.myInitialCompliance;
            results.initialStress =  obj.myInitialStress;
            results.initialFem = obj.myInitialFem;
            results.flag = flag;
            results.finalParams = optimalParams;
            results.finalVolume  = fem.computeArea();
            results.finalCompliance = fem.computeCompliance();
            results.finalStress = fem.myMaxStress;
            results.finalFEM = fem;
            results.nFEAs = output.funcCount;
        end
    end
end
