classdef cantilever2DMinVolumeFD 
    properties(GetAccess = 'public', SetAccess = 'private')
        myInitialShapeParams; % used for scaling
        myUB; % upper bound on the shape parameters
        myLB;% lower bound on the shape parameters
        myInitialVolume; % used for scaling volume
        myInitialCompliance; % used for scaling compliance
        myNumElems;
        myFiniteDifferenceStepSize;
        myTerminationTolerance;
        myInitialFem; % Initial Fem structure
    end
    methods
        function obj = cantilever2DMinVolumeFD(params0,nElements)
            obj.myNumElems = nElements;
            obj.myInitialShapeParams = params0;
            fem = obj.solveCantileverProblem(params0);
            obj.myInitialFem = fem;
            obj.myInitialCompliance = fem.computeCompliance();
            obj.myInitialVolume = fem.computeArea();
            obj.myFiniteDifferenceStepSize = 1e-6; %default
            obj.myTerminationTolerance = 1e-6;% default
        end
        function fem = solveCantileverProblem(obj,params)
            fem = obj.createCantileverProblem(params);
            fem = fem.assemble();
            fem = fem.solve();
        end
        function fem = createCantileverProblem(obj,params)
            cantileverBeam = obj.createGeom(params);
            fem = TriElasticity(cantileverBeam,obj.myNumElems);
            fem = fem.setYoungsModulus(3.05e7);
            fem = fem.setPoissonsRatio(0.28);
            fem = fem.fixEdge([2,10]);
            fem = fem.applyYForceOnEdge(6,-1000);
        end
        
        function [geom] = createGeom(~,params)
            L = 2.0;%length
            H = 1;%height
            h = 0.2; % length of force edge
            a = params(1); % corner cutouts
            b = params(2); % left edge cutout
            geom.vertices = [b 0;0 -b; 0 -H/2;L-a -H/2;L -H/2+a; L -h/2; L h/2; L H/2-a; L-a H/2; 0 H/2; 0 b ]';
            geom.edges = [1 2; 2 3; 3 4; 4 5; 5 6; 6 7; 7 8; 8 9; 9 10; 10 11; 11 1];
        end
        function obj = setParamsUpperBound(obj,paramsUpper)
            obj.myUB = paramsUpper;
        end
        function obj = setTerminationTolerance(obj,tol)
            obj.myTerminationTolerance = tol;
        end
        function obj = setFiniteDifferenceStepSize(obj,step)
            obj.myFiniteDifferenceStepSize = step;
        end
        function obj = setParamsLowerBound(obj,paramsLower)
            obj.myLB = paramsLower;
        end   
        function [cineq,ceq] = complianceConstraint(obj,x)
            params = x.*obj.myInitialShapeParams;
            fem = obj.solveCantileverProblem(params);
            compliance = fem.computeCompliance();
            ceq = [];
            cineq = compliance/obj.myInitialCompliance -1; 
        end  
        function f = volumeObjective(obj,x)
            params = x.*obj.myInitialShapeParams;
            fem = obj.createCantileverProblem(params);
            currentVolume = fem.computeArea();
            f = currentVolume/(obj.myInitialVolume);
        end
        function [results] = optimize(obj) 
            LB = obj.myLB./obj.myInitialShapeParams;
            UB = obj.myUB./obj.myInitialShapeParams;
            x0 = ones(1,2); 
            opt = optimoptions('fmincon',...
            'TolX',obj.myTerminationTolerance,...
            'TolFun',obj.myTerminationTolerance,...
            'ConstraintTolerance',obj.myTerminationTolerance,...
            'FiniteDifferenceStepSize',obj.myFiniteDifferenceStepSize);        
            [xMin,fMin,flag,output] = fmincon(@obj.volumeObjective,x0,[],[],...
                [],[],LB,UB,@obj.complianceConstraint,opt);
            optimalParams = xMin.*obj.myInitialShapeParams;
            fem = obj.solveCantileverProblem(optimalParams); 
            results.initialFem = obj.myInitialFem;
            results.initialParams =  obj.myInitialShapeParams;
            results.initialVolume =  obj.myInitialVolume;
            results.initialCompliance =  obj.myInitialCompliance;
            results.flag = flag;
            results.finalParams = optimalParams;
            results.finalVolume  = fem.computeArea();
            results.finalCompliance = fem.computeCompliance();
            results.finalFEM = fem;
            % each function call involves additional FEAs per parameter
            results.nFEAs = output.funcCount*(1+length(obj.myInitialShapeParams));
        end
    end
end

