classdef testFunctionsND
    % This class encapsulates numerous N-D test functions
    methods(Static)
        function [f,g,h] = f1(xVec)
            % returns  x^2-x*y+y^2-x-2*y and its derivatives
            x = xVec(1);y = xVec(2);
            f = x.^2 - x.*y + y.^2 -x -2*y;
            if (nargout > 1), g = [2*x-y-1; -x+2*y-2];end
            if (nargout > 2), h = [2 -1; -1 2];end
        end
        function [g,h] = g1(xVec)
            % returns the gradient and hessian of x^2-x*y+y^2-x-2*y 
            x = xVec(1);y = xVec(2);
            g = [2*x-y-1; -x+2*y-2];
            if (nargout > 1), h = [2 -1; -1 2];end
        end
        function [f,g,h] = f2(xVec)
            % returns (x-1).^4 + (y-1).^2 and its derivatives
            x = xVec(1);y = xVec(2);
            f = (x-1).^4 + (y-1).^2;
            if (nargout > 1), g = [4*(x-1).^3; 2*(y-1)];end
            if (nargout > 2), h =[12*(x-1).^2 0; 0 2];end
        end
        function [f,g,h] = f3(xVec)
            % returns x^2 +4*x*y + y^2 and its derivatives
            x = xVec(1);y = xVec(2);
            f = x.^2 +4*x.*y + y.^2;
            if (nargout > 1), g = [2*x+4*y; 4*x+2*y];end
            if (nargout > 2), h = [2 4; 4 2];end
        end
        function [f,g,h] = f4(xVec)
            % returns 2-D Rosenbrock: 100*(y-x^2)^2 + (1-x)^2 and its derivatives
            x = xVec(1);y = xVec(2);
            f = 100*(y-x^2)^2 + (1-x)^2;
            if (nargout > 1)
                g = [2*x-400*x*(-x^2+y)-2; -200*x^2+200*y];
            end
            if (nargout > 2), h = [1200*x^2-400*y+2 -400*x;-400*x 200];end
        end
        function g = g4(xVec)
            % returns the gradient of the 2-D Rosenbrock: 100*(y-x^2)^2 + (1-x)^2
            x = xVec(1);y = xVec(2);
            g = [2*x-400*x*(-x^2+y)-2; -200*x^2+200*y];
        end
        function [f] = f5(xVec)
            % returns abs(x-1) + (y-1)^2
            x = xVec(1);y = xVec(2);
            f = abs(x-1) + (y-1)^2;
        end
        function [f,g,h] = f6(x)
           % returns N-D Rosenbrock: sum(100*(x(i+1)-x(i)^2)^2 + (1-x(i))^2) and its derivatives
            M = numel(x);
            N = M-1;f = 0;
            for i = 1:N, f = f + 100*(x(i+1)-x(i)^2)^2 + (1-x(i))^2; end
            if (nargout > 1)
                g = zeros(M,1);
                g(1) = -400*(x(2)-x(1)^2)*x(1)-2*(1-x(1));
                for i = 2:N
                    g(i) =  200*(x(i)-x(i-1)^2) - 400*(x(i+1)-x(i)^2)*x(i)-2*(1-x(i));
                end
                g(N+1) = 200*(x(N+1)-x(N)^2);
            end
            if (nargout > 2)
                h = zeros(N+1,N+1);
                for i = 1:N
                    h(i,i) = h(i,i) + 1200*x(i).^2-400*x(i+1)+2;
                    h(i,i+1) = h(i,i+1) + (-400*x(i));
                    h(i+1,i) = h(i,i+1);
                    h(i+1,i+1) = h(i+1,i+1) + 200;               
                end
            end
        end
         function [g] = g6(x)
           % returns the gradient of N-D Rosenbrock: sum(100*(x(i+1)-x(i)^2)^2 + (1-x(i))^2) and its derivatives
           M = numel(x);
           N = M-1;
           g = zeros(M,1);
           g(1) = -400*(x(2)-x(1)^2)*x(1)-2*(1-x(1));
           for i = 2:N
               g(i) =  200*(x(i)-x(i-1)^2) - 400*(x(i+1)-x(i)^2)*x(i)-2*(1-x(i));
           end
           g(N+1) = 200*(x(N+1)-x(N)^2);
         end
         function [f,g,h] = f7(xVec)
            % returns  3*x^2 - x*y + 4*y^2 -x - 2*y + 10 and its derivatives
            x = xVec(1);y = xVec(2);
            f = 3*x^2 - x*y + 4*y^2 -x - 2*y + 10;
            if (nargout > 1), g  = [6*x-y-1;-x+8*y-2];end
            if (nargout > 2), h =[6 -1;-1 8];end
        end 
        function [g,h] = g7(xVec)
            % returns  gradient and hessian of 3*x^2 - x*y + 4*y^2 -x - 2*y + 10
            x = xVec(1);y = xVec(2);
            g  = [6*x-y-1;-x+8*y-2];
            if (nargout > 2), h =[6 -1;-1 8];end
        end 
        function [f,g,h] = f8(xVec)
            % returns  3*x^2 + 4*y^2 -x - 2*y + 10 and its derivatives
            x = xVec(1);y = xVec(2);
            f = 3*x^2  + 4*y^2 -x - 2*y + 10;
            if (nargout > 1), g = [6*x-1;8*y-2];end
            if (nargout > 2), h =[6 0;0 8];end
        end 
        function [f] = f9(xVec)
            % returns  x.*x -0.2*x.*x.*y + y.*y - x -2*y
            x = xVec(1);y = xVec(2);
            f = x*x -0.2*x*x*y + y*y - x -2*y;
        end
        function [f,g,h] = f10(xVec) 
            %  potential energy of a 2-spring system. 
            u = xVec(1);v = xVec(2);
            L12 = sqrt(u*u+(1+v)^2);
            L13 = sqrt(u^2+(1-v)^2);
            f = 0.5*(100*(L12 -1)^2 + 50*(L13-1)^2) - (10*u+8*v);
            % gradient 
            g(1,1) = 100*(L12-1)*u/L12 + 50*(L13-1)*u/L13 -10;
            g(2,1) = 100*(L12-1)*(1+v)/L12 - 50*(L13-1)*(1-v)/L13 -8;
            
            h(1,1) = (50*((v - 1)^2 + u^2)^(1/2) - 50)/((v - 1)^2 + u^2)^(1/2) + (100*((v + 1)^2 + u^2)^(1/2) - 100)/((v + 1)^2 + u^2)^(1/2) + (50*u^2)/((v - 1)^2 + u^2) + (100*u^2)/((v + 1)^2 + u^2) - (u^2*(50*((v - 1)^2 + u^2)^(1/2) - 50))/((v - 1)^2 + u^2)^(3/2) - (u^2*(100*((v + 1)^2 + u^2)^(1/2) - 100))/((v + 1)^2 + u^2)^(3/2);
            h(1,2) = (25*u*(2*v - 2))/((v - 1)^2 + u^2) + (50*u*(2*v + 2))/((v + 1)^2 + u^2) - (u*(2*v - 2)*(50*((v - 1)^2 + u^2)^(1/2) - 50))/(2*((v - 1)^2 + u^2)^(3/2)) - (u*(2*v + 2)*(100*((v + 1)^2 + u^2)^(1/2) - 100))/(2*((v + 1)^2 + u^2)^(3/2));      
            h(2,1) = h(1,2);
            h(2,2) = (50*((v - 1)^2 + u^2)^(1/2) - 50)/((v - 1)^2 + u^2)^(1/2) + (100*((v + 1)^2 + u^2)^(1/2) - 100)/((v + 1)^2 + u^2)^(1/2) + (25*(2*v - 2)*(v - 1))/((v - 1)^2 + u^2) + (50*(2*v + 2)*(v + 1))/((v + 1)^2 + u^2) - ((2*v - 2)*(50*((v - 1)^2 + u^2)^(1/2) - 50)*(v - 1))/(2*((v - 1)^2 + u^2)^(3/2)) - ((2*v + 2)*(100*((v + 1)^2 + u^2)^(1/2) - 100)*(v + 1))/(2*((v + 1)^2 + u^2)^(3/2));
        end
        function [g] = g10(xVec)
            % the gradient of f10
            u = xVec(1);v = xVec(2);
            L12 = sqrt(u*u+(1+v)^2);
            L13 = sqrt(u^2+(1-v)^2);
            % gradient 
            g(1,1) = 100*(L12-1)*u/L12 + 50*(L13-1)*u/L13 -10;
            g(2,1) = 100*(L12-1)*(1+v)/L12 - 50*(L13-1)*(1-v)/L13 -8;
        end
        function [f] = f11(xVec) % non-smooth function
            x = xVec(1);y = xVec(2);
            f = abs(x-2) + abs(y-3);
        end
        function [f,g,h] = f12(xVec) % Quadratic 3-variable
            % f(x,y,z)  = x^2- x*y +y^2 + 5*z^2  + x*z -3*y*z -x-2*y+4*z
            x = xVec(1);y = xVec(2);z = xVec(3);
            f = x^2- x*y +y^2 + 5*z^2  + x*z -3*y*z -x-2*y+4*z;
            if (nargout > 1), g = [2*x - y + z - 1;2*y - x - 3*z - 2;x - 3*y + 10*z + 4];end
            if (nargout > 2), h =[2 -1 1;-1 2 -3;1 -3 10];end
        end
        function [f] = f13(xVec) % Bird function
            x = xVec(1);y = xVec(2);
            f = sin(x).*exp((1-cos(y)).^2)+cos(y).*exp((1-sin(x)).^ 2)+(x-y).^2;
        end
    end
end
